/*
  *  
  *	see tcpha_be.c for infoes about TCPHA
  *
  */

#include "tcpha_be.h"
#include "tcpha_be_misc.h"
#include "tcpha_be_con.h"

#ifdef	CONFIG_TCPHA_BE_DEBUG
int tcpha_be_connmagic=0;
int tcpha_be_ctrlconnmagic=0;
#define tf_set_connmagic(x) 	x->magic=tcpha_be_connmagic++
#define tf_set_ctrlconnmagic(x)	x->magic=tcpha_be_ctrlconnmagic++;
#else	/* NO DEBUGGING at ALL */
#define tf_set_connmagic(x) do {} while (0)
#define tf_set_ctrlconnmagic(x)  do {} while (0)
#endif

typedef enum ha_ret {
	TCP_HA_OK=0,
	TCP_HA_NO
}ha_ret_t;

typedef struct tcp_ha_ret {
	__u32 magic;
	__u32 connmagic;
	ha_ret_t msg;
}tcp_ha_ret_t;

struct tcpha_be_conn *
tcpha_be_conn_create(struct socket *sock, struct tcpha_be_service *svc)
{
	struct tcpha_be_conn *cp;

	cp = kmem_cache_alloc(tcpha_be_conn_cachep, GFP_KERNEL);
	if (!cp) {
		ERR_RL("create_conn no memory available\n");
		return NULL;
	}
	memset(cp, 0, sizeof(*cp));
	/* clone the socket */
	cp->fsock = sock_alloc();
	if (!cp->fsock) {
		kmem_cache_free(tcpha_be_conn_cachep, cp);
		return NULL;
	}
	tf_set_connmagic(cp);

	cp->fsock->type = sock->type;
	cp->fsock->ops = sock->ops;

	return cp;
}

static inline struct sk_buff *tcpha_be_conn_getinfo(struct socket *sock,struct tcpha_conn_info **info)
{
	struct sk_buff *skb;
	int len;
	
	skb = skb_dequeue(&sock->sk->receive_queue);
	if (!skb)
		goto out_error;
	len = skb->len;
	if (!pskb_may_pull(skb, TCPHACONNINFOSIZE)||!skb->data||*(__u32 *)skb->data!=TCPHAMAGIC) {
		kfree_skb(skb);
		goto out_error;
	}
	*info=(struct tcpha_conn_info *)skb->data;
	skb->data+=TCPHACONNINFOSIZE;
	skb->len-=TCPHACONNINFOSIZE;
	return skb;
	out_error:
	return NULL;
}

int
redirect_to_local(struct tcpha_be_conn *,struct tcpha_conn_info *,struct sk_buff *skb, __u32, __u16);

int
tcpha_be_conn_handle(struct tcpha_be_conn *conn, struct tcpha_be_service *svc)
{
	struct socket *fsock;
	struct tcpha_conn_info *info;
	struct sk_buff *skb;
	tcp_ha_ret_t ret;
	
	fsock= conn->fsock;
	if (fsock->sk->state != TCP_ESTABLISHED) {
		if (fsock->sk->state == TCP_CLOSE_WAIT) {
			tcpha_be_conn_release(conn);
			return 0;
		}
		DBG_RL("Error control connection not established (state %d)\n",
			   fsock->sk->state);
		goto out_error;
	}

	DBG("server:%u.%u.%u.%u:%d created control connection:%d by process:%d\n",
		       NIPQUAD(tbs.conf.raddr),ntohs(tbs.conf.port), conn->magic,current->pid);
	again:
	while (skb_queue_empty(&(fsock->sk->receive_queue))) {
		if (!interruptible_sleep_on_timeout(&fsock->wait, 15*60*HZ)) {
			DBG("control connection:%d time out\n",conn->magic);
			goto out_error;
		}
		if (signal_pending(current)&&skb_queue_empty(&(fsock->sk->receive_queue))) {
			DBG("received signal by process:%d\n",current->pid);
			goto out;
		}
	}

	if (!(skb=tcpha_be_conn_getinfo(fsock,&info))) {
		DBG_RL("error reading handoff request from FE\n");
		goto out_error;
	}
	ret.connmagic=info->connmagic;
	ret.msg=redirect_to_local(conn,info,skb,tbs.conf.redirect_addr,tbs.conf.redirect_port)?TCP_HA_NO:TCP_HA_OK;
	ret.magic=TCPHAMAGIC;
	if (tcpha_be_sendbuffer(fsock,(char *)&ret,sizeof(ret),0)!=sizeof(ret)) {
		DBG_RL("send handoff response error\n");
		goto out_error;
	}
	if (ret.msg==TCP_HA_OK)
		goto again;
	else
		goto out_error;
	out:
	return 0;
	out_error:
	return -1;
}

void
tcpha_be_conn_release(struct tcpha_be_conn *conn)
{
	/* close the connetion with FE */
	if (conn->fsock) {
		DBG("close control connection:%d\n",conn->magic);
		sock_release(conn->fsock);
	}
	kmem_cache_free(tcpha_be_conn_cachep, conn);
}

